// SPDX-License-Identifier: Apache-2.0 AND GPL-3.0-or-later

package io.github.muntashirakon.AppManager.scanner;

import android.app.Activity;
import android.content.Intent;
import android.graphics.Typeface;
import android.os.Bundle;
import android.view.LayoutInflater;
import android.view.Menu;
import android.view.MenuInflater;
import android.view.MenuItem;
import android.view.View;
import android.view.ViewGroup;
import android.widget.Filter;
import android.widget.Filterable;
import android.widget.TextView;

import androidx.annotation.NonNull;
import androidx.annotation.Nullable;
import androidx.annotation.UiThread;
import androidx.fragment.app.Fragment;
import androidx.lifecycle.ViewModelProvider;
import androidx.recyclerview.widget.LinearLayoutManager;

import com.google.android.material.card.MaterialCardView;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Locale;

import io.github.muntashirakon.AppManager.R;
import io.github.muntashirakon.AppManager.misc.AdvancedSearchView;
import io.github.muntashirakon.AppManager.utils.TextUtilsCompat;
import io.github.muntashirakon.AppManager.utils.UIUtils;
import io.github.muntashirakon.AppManager.utils.appearance.ColorCodes;
import io.github.muntashirakon.util.UiUtils;
import io.github.muntashirakon.widget.RecyclerView;

import static io.github.muntashirakon.AppManager.misc.AdvancedSearchView.SEARCH_TYPE_REGEX;

public class ClassListingFragment extends Fragment implements AdvancedSearchView.OnQueryTextListener {
    private TextView mEmptyView;
    private boolean mTrackerClassesOnly;
    private ClassListingAdapter mClassListingAdapter;

    private List<String> mAllClasses;
    private List<String> mTrackerClasses;
    private ScannerViewModel mViewModel;
    private ScannerActivity mActivity;

    public ClassListingFragment() {
        setHasOptionsMenu(true);
    }

    @Nullable
    @Override
    public View onCreateView(@NonNull LayoutInflater inflater, @Nullable ViewGroup container, @Nullable Bundle savedInstanceState) {
        return inflater.inflate(R.layout.fragment_class_lister, container, false);
    }

    @Override
    public void onViewCreated(@NonNull View view, @Nullable Bundle savedInstanceState) {
        mViewModel = new ViewModelProvider(requireActivity()).get(ScannerViewModel.class);
        mActivity = (ScannerActivity) requireActivity();
        mAllClasses = mViewModel.getAllClasses();
        mTrackerClasses = mViewModel.getTrackerClasses();
        if (mAllClasses == null) {
            mActivity.onBackPressed();
            return;
        }
        if (mTrackerClasses == null) {
            mTrackerClasses = Collections.emptyList();
        }

        mTrackerClassesOnly = false;

        RecyclerView listView = view.findViewById(R.id.list_item);
        UiUtils.applyWindowInsetsAsPaddingNoTop(listView);
        mEmptyView = view.findViewById(android.R.id.empty);
        listView.setEmptyView(mEmptyView);
        mClassListingAdapter = new ClassListingAdapter(mActivity, mViewModel);
        listView.setLayoutManager(new LinearLayoutManager(mActivity));
        listView.setAdapter(mClassListingAdapter);
        showProgress(true);
        setAdapterList();
    }

    @Override
    public void onResume() {
        super.onResume();
        if (mClassListingAdapter != null && !TextUtilsCompat.isEmpty(mClassListingAdapter.mConstraint)) {
            mClassListingAdapter.filter();
        }
    }

    @UiThread
    private void setAdapterList() {
        if (!mTrackerClassesOnly) {
            mClassListingAdapter.setDefaultList(mTrackerClasses);
            mActivity.setSubtitle(getString(R.string.tracker_classes));
        } else {
            mClassListingAdapter.setDefaultList(mAllClasses);
            mActivity.setSubtitle(getString(R.string.all_classes));
        }
        showProgress(false);
    }

    @Override
    public boolean onQueryTextChange(String newText, @AdvancedSearchView.SearchType int type) {
        if (mClassListingAdapter != null) {
            mClassListingAdapter.filter(newText, type);
        }
        return true;
    }

    @Override
    public boolean onQueryTextSubmit(String query, int type) {
        return false;
    }

    @Override
    public void onCreateOptionsMenu(@NonNull Menu menu, @NonNull MenuInflater inflater) {
        inflater.inflate(R.menu.fragment_class_lister_actions, menu);
        AdvancedSearchView searchView = (AdvancedSearchView) menu.findItem(R.id.action_search).getActionView();
        searchView.setOnQueryTextListener(this);
        super.onCreateOptionsMenu(menu, inflater);
    }

    @Override
    public boolean onOptionsItemSelected(@NonNull MenuItem item) {
        int id = item.getItemId();
        if (id == R.id.action_toggle_class_listing) {
            mTrackerClassesOnly = !mTrackerClassesOnly;
            setAdapterList();
        } else return super.onOptionsItemSelected(item);
        return true;
    }

    private void showProgress(boolean willShow) {
        mActivity.showProgress(willShow);
        mEmptyView.setText(willShow ? R.string.loading : R.string.no_tracker_class);
    }

    static class ClassListingAdapter extends RecyclerView.Adapter<ClassListingAdapter.ViewHolder> implements Filterable {
        private Filter mFilter;
        private String mConstraint;
        @AdvancedSearchView.SearchType
        private int mFilterType = AdvancedSearchView.SEARCH_TYPE_CONTAINS;
        private List<String> mDefaultList;
        private final Activity mActivity;
        private final ScannerViewModel mViewModel;
        @NonNull
        private final List<String> mAdapterList = new ArrayList<>();
        private final int mCardColor0;
        private final int mCardColor1;
        private final int mQueryStringHighlightColor;

        ClassListingAdapter(@NonNull Activity activity, @NonNull ScannerViewModel viewModel) {
            mActivity = activity;
            mViewModel = viewModel;
            mCardColor0 = ColorCodes.getListItemColor0(activity);
            mCardColor1 = ColorCodes.getListItemColor1(activity);
            mQueryStringHighlightColor = ColorCodes.getQueryStringHighlightColor(activity);
        }

        @UiThread
        void setDefaultList(@NonNull List<String> list) {
            synchronized (mAdapterList) {
                mAdapterList.clear();
                mAdapterList.addAll(list);
            }
            mDefaultList = list;
            filter();
            notifyDataSetChanged();
        }

        void filter() {
            if (!TextUtilsCompat.isEmpty(mConstraint)) {
                filter(mConstraint, mFilterType);
            }
        }

        void filter(String query, @AdvancedSearchView.SearchType int filterType) {
            mConstraint = query;
            mFilterType = filterType;
            getFilter().filter(mConstraint);
        }

        @Override
        public int getItemCount() {
            synchronized (mAdapterList) {
                return mAdapterList.size();
            }
        }

        @Override
        public long getItemId(int position) {
            synchronized (mAdapterList) {
                return mDefaultList.indexOf(mAdapterList.get(position));
            }
        }

        @NonNull
        @Override
        public ViewHolder onCreateViewHolder(@NonNull ViewGroup parent, int viewType) {
            View view = LayoutInflater.from(parent.getContext()).inflate(io.github.muntashirakon.ui.R.layout.m3_preference, parent, false);
            return new ViewHolder(view);
        }

        @Override
        public void onBindViewHolder(@NonNull ViewHolder holder, int position) {
            String className;
            synchronized (mAdapterList) {
                className = mAdapterList.get(position);
            }
            TextView textView = holder.classNameView;
            textView.setTypeface(Typeface.MONOSPACE);
            if (mConstraint != null && className.toLowerCase(Locale.ROOT).contains(mConstraint)) {
                // Highlight searched query
                textView.setText(UIUtils.getHighlightedText(className, mConstraint, mQueryStringHighlightColor));
            } else {
                textView.setText(className);
            }
            holder.itemView.setCardBackgroundColor(position % 2 == 0 ? mCardColor1 : mCardColor0);
            holder.itemView.setOnClickListener(v -> {
                try {
                    Intent intent = new Intent(mActivity, ClassViewerActivity.class);
                    intent.putExtra(ClassViewerActivity.EXTRA_URI, mViewModel.getUriFromClassName(className));
                    intent.putExtra(ClassViewerActivity.EXTRA_APP_NAME, mActivity.getTitle());
                    mActivity.startActivity(intent);
                } catch (Exception e) {
                    e.printStackTrace();
                    UIUtils.displayLongToast(e.toString());
                }
            });
        }

        @Override
        public Filter getFilter() {
            if (mFilter == null)
                mFilter = new Filter() {
                    @Override
                    protected FilterResults performFiltering(CharSequence charSequence) {
                        String constraint = mFilterType == SEARCH_TYPE_REGEX ? charSequence.toString()
                                : charSequence.toString().toLowerCase(Locale.ROOT);
                        FilterResults filterResults = new FilterResults();
                        if (constraint.length() == 0) {
                            filterResults.count = 0;
                            filterResults.values = null;
                            return filterResults;
                        }

                        List<String> list = AdvancedSearchView.matches(
                                constraint,
                                mDefaultList,
                                (AdvancedSearchView.ChoiceGenerator<String>) object -> mFilterType == SEARCH_TYPE_REGEX ? object
                                        : object.toLowerCase(Locale.ROOT),
                                mFilterType);

                        filterResults.count = list.size();
                        filterResults.values = list;
                        return filterResults;
                    }

                    @Override
                    protected void publishResults(CharSequence charSequence, FilterResults filterResults) {
                        synchronized (mAdapterList) {
                            mAdapterList.clear();
                            if (filterResults.values == null) {
                                mAdapterList.addAll(mDefaultList);
                            } else {
                                //noinspection unchecked
                                mAdapterList.addAll((List<String>) filterResults.values);
                            }
                            notifyDataSetChanged();
                        }
                    }
                };
            return mFilter;
        }

        public static class ViewHolder extends RecyclerView.ViewHolder {
            final MaterialCardView itemView;
            final TextView classNameView;

            public ViewHolder(@NonNull View itemView) {
                super(itemView);
                this.itemView = (MaterialCardView) itemView;
                itemView.findViewById(android.R.id.title).setVisibility(View.GONE);
                itemView.findViewById(R.id.icon_frame).setVisibility(View.GONE);
                classNameView = itemView.findViewById(android.R.id.summary);
            }
        }
    }
}
